
import logging
import netaddr

from pytos.secureapp.xml_objects.base_types import Base_Link_Target, Base_Object, URL_Link, Network_Object, Service_Object
from pytos.common.base_types import XML_Object_Base, XML_List, Group_Service_Type, Single_Service_Type, \
    Range_Service_Type, Any_Service_Type, Comparable
from pytos.common.definitions.xml_tags import Attributes, Elements
from pytos.common.functions import str_to_bool, XML_LOGGER_NAME
from pytos.common.functions.xml import get_xml_text_value, get_xml_int_value, get_xml_node

logger = logging.getLogger(XML_LOGGER_NAME)


class Applications_List(XML_List):
    def __init__(self, applications):
        """
        :type: applications: list[Application]
        """
        super().__init__(Elements.APPLICATIONS, applications)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        applications = []
        for application_node in xml_node.iter(tag=Elements.APPLICATION):
            applications.append(Application.from_xml_node(application_node))
        return cls(applications)


class Application(XML_Object_Base, Comparable):
    def __init__(self, app_id, name, comment, decommissioned, owner, editors, created, modified, status, connections,
                 open_tickets, customer=None, connection_to_application_packs=None):
        self.id = app_id
        self.name = name
        self.comment = comment
        self.decommissioned = decommissioned
        self.owner = owner
        self.editors = editors
        self.created = created
        self.modified = modified
        self.status = status
        self.connections = connections
        self.open_tickets = open_tickets
        # customer is used only on >=15.1 but if send on prev version it will be ignored and no error will be generated
        self.customer = customer

        self.connection_to_application_packs = connection_to_application_packs
        super().__init__(Elements.APPLICATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        app_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        decommissioned = get_xml_text_value(xml_node, Elements.DECOMMISSIONED)
        owner = Application_Owner.from_xml_node(get_xml_node(xml_node, Elements.OWNER))
        created = get_xml_text_value(xml_node, Elements.CREATED)
        modified = get_xml_text_value(xml_node, Elements.MODIFIED)
        status = get_xml_text_value(xml_node, Elements.STATUS)
        editors = XML_List.from_xml_node_by_tags(xml_node, Elements.EDITORS, Elements.EDITOR, Application_Editor)
        open_tickets = XML_List.from_xml_node_by_tags(xml_node, Elements.OPEN_TICKETS, Elements.TICKET,
                                                      Application_Open_Ticket)
        # For <15.1 compatibility we would not force to get this one as it will fail
        customer_xml_node = get_xml_node(xml_node, Elements.CUSTOMER, True)
        if customer_xml_node is not None:
            customer = Customer.from_xml_node(customer_xml_node)
        else:
            customer = None

        connection_to_application_packs = XML_List.from_xml_node_by_tags(xml_node,
                                                                         Elements.CONNECTION_TO_APPLICATION_PACKS,
                                                                         Elements.CONNECTION_TO_APPLICATION_PACK,
                                                                         Connection_To_Application_Pack, True)

        connections = []
        for connection_node in xml_node.iter(tag=Elements.CONNECTION):
            connections.append(Application_Connection.from_xml_node(connection_node))
        return cls(app_id, name, comment, decommissioned, owner, editors, created, modified, status, connections,
                   open_tickets, customer, connection_to_application_packs)

    def is_decommissioned(self):
        return str_to_bool(self.decommissioned)

    def _key(self):
        return self.id, self.name

    def __repr__(self):
        return 'Application. Name: {}, Owner: {}'.format(self.name, self.owner.display_name)

    def __str__(self):
        return repr(self)


class Application_Reference(Base_Link_Target):
    def __init__(self, app_id, display_name, name, link):
        super().__init__(Elements.APPLICATION, app_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """

        app_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(app_id, display_name, name, link)


class Application_Open_Ticket(Base_Link_Target):
    def __init__(self, ticket_id, display_name, name, link):
        super().__init__(Elements.TICKET, ticket_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """

        ticket_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(ticket_id, display_name, name, link)


class Application_Editor(Base_Link_Target):
    def __init__(self, editor_id, display_name, name, link):
        super().__init__(Elements.EDITOR, editor_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """

        editor_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(editor_id, display_name, name, link)


class Application_Owner(Base_Link_Target):
    def __init__(self, owner_id, display_name, name, link):
        super().__init__(Elements.OWNER, owner_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """

        owner_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(owner_id, display_name, name, link)


class Server(Base_Link_Target):
    def __init__(self, server_id, display_name, name, server_type, link):
        self.type = server_type
        super().__init__(Elements.SERVER, server_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        server_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        server_type = get_xml_text_value(xml_node, Elements.TYPE)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(server_id, display_name, name, server_type, link)


class Connection_To_Application(Base_Link_Target):
    def __init__(self, conn_to_app_id, display_name, name, link):
        super().__init__(Elements.CONNECTION_TO_APPLICATION, conn_to_app_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        conn_to_app_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(conn_to_app_id, display_name, name, link)


class Connection_To_Application_Pack(Base_Link_Target):
    def __init__(self, conn_to_app_pack_id, display_name, name, link):
        super().__init__(Elements.CONNECTION_TO_APPLICATION_PACK, conn_to_app_pack_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        conn_to_app_pack_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(conn_to_app_pack_id, display_name, name, link)


class Application_Pack(Base_Link_Target):
    def __init__(self, app_pack_id, display_name, name, link):
        super().__init__(Elements.APPLICATION_PACK, app_pack_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        app_pack_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(app_pack_id, display_name, name, link)


class Application_Connection(Base_Link_Target):
    def __init__(self, connection_id, display_name, name, link):
        super().__init__(Elements.CONNECTION, connection_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(connection_id, display_name, name, link)


class Tag_Reference(Base_Link_Target):
    def __init__(self, tag_id, display_name, name, link):
        super().__init__(Elements.TAG, tag_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        tag_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(tag_id, display_name, name, link)


class Tag_Servers(XML_List):
    """
    :type tags: list[Tag_Reference]
    """

    def __init__(self, tags):
        super().__init__(Elements.TAG_SERVERS, tags)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        tags = []
        for connection_node in xml_node.iter(tag=Elements.TAG):
            tags.append(Tag_Reference.from_xml_node(connection_node))
        return cls(tags)


class User_List(XML_List):
    """
    :type users: list[User]
    """

    def __init__(self, users):
        super().__init__(Elements.USERS, users)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        users = []
        for user_node in xml_node.iter(tag=Elements.USER):
            users.append(User.from_xml_node(user_node))
        return cls(users)


class User(Base_Object):
    def __init__(self, display_name, is_global, user_id, name, user_type, ip):
        self.ip = ip
        super().__init__(Elements.USER, display_name, is_global, user_id, name, user_type)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        user_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        user_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, Elements.IP)
        return cls(display_name, is_global, user_id, name, user_type, ip)

    def _key(self):
        return self.ip,


class Connection_List(XML_List):
    """
    :type connections: list[Detailed_Application_Connection]
    """

    def __init__(self, connections):
        super().__init__(Elements.CONNECTIONS, connections)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connections = []
        for connection_node in xml_node.iter(tag=Elements.CONNECTION):
            connections.append(Detailed_Application_Connection.from_xml_node(connection_node))
        return cls(connections)


class ConnectionExtendedList(XML_List):
    def __init__(self, connections):
        super().__init__(Elements.CONNECTIONS_EXTENDED, connections)

    @classmethod
    def from_xml_node(cls, xml_node):
        connections = []
        for con_node in xml_node.iter(tag=Elements.CONNECTION_EXTENDED):
            connections.append(ConnectionExtended.from_xml_node(con_node))
        return cls(connections)


class Detailed_Connection_To_Application(XML_Object_Base):
    def __init__(self, con_to_app_id, name, comment, application_id, application_interface_id, connections):
        self.id = con_to_app_id
        self.name = name
        self.comment = comment
        self.application_id = application_id
        self.application_interface_id = application_interface_id
        self.connections = connections
        super().__init__(Elements.CONNECTION_TO_APPLICATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        con_to_app_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        app_interface_id = get_xml_int_value(xml_node, Elements.APPLICATION_INTERFACE_ID)
        connections = XML_List.from_xml_node_by_tags(xml_node, Elements.CONNECTIONS, Elements.CONNECTION,
                                                     Detailed_Application_Connection, True)
        return cls(con_to_app_id, name, comment, application_id, app_interface_id, connections)


class Detailed_Connection_To_Application_Pack(XML_Object_Base):
    def __init__(self, conn_to_app_pack_id, name, comment, created, modified, application, application_pack,
                 connection_to_applications, tags_servers):
        self.id = conn_to_app_pack_id
        self.name = name
        self.comment = comment
        self.created = created
        self.modified = modified
        self.application = application
        self.application_pack = application_pack
        self.connection_to_applications = connection_to_applications
        self.tags_servers = tags_servers
        super().__init__(Elements.CONNECTION_TO_APPLICATION_PACK)

    @classmethod
    def from_xml_node(cls, xml_node):
        conn_to_app_pack_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        created = get_xml_text_value(xml_node, Elements.CREATED)
        modified = get_xml_text_value(xml_node, Elements.MODIFIED)
        app = None
        app_node = get_xml_node(xml_node, Elements.APPLICATION)
        if app_node:
            app = Application_Reference.from_xml_node(app_node)
        application_pack = None
        app_pack_node = get_xml_node(xml_node, Elements.APPLICATION_PACK)
        if app_pack_node:
            Application_Pack.from_xml_node(app_pack_node)
        connection_to_applications = XML_List.from_xml_node_by_tags(xml_node, Elements.CONNECTION_TO_APPLICATIONS,
                                                                    Elements.CONNECTION_TO_APPLICATION,
                                                                    Connection_To_Application, True)
        tags_servers = XML_List.from_xml_node_by_tags(xml_node, Elements.TAGS_SERVERS, Elements.TAG_SERVERS,
                                                      Tag_Servers, True)
        return cls(conn_to_app_pack_id, name, comment, created, modified, app, application_pack,
                   connection_to_applications, tags_servers)


class Connection_To_Application_Packs(XML_List):
    """
    :type connection_to_application_packs: list[Detailed_Connection_To_Application_Pack]
    """

    def __init__(self, connection_to_application_packs):
        super().__init__(Elements.CONNECTION_TO_APPLICATION_PACKS, connection_to_application_packs)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connection_to_application_packs = []
        for con_to_app_node in xml_node.iter(tag=Elements.CONNECTION_TO_APPLICATION_PACK):
            connection_to_application_packs.append(
                    Detailed_Connection_To_Application_Pack.from_xml_node(con_to_app_node))
        return cls(connection_to_application_packs)


class Connections_To_Applications(XML_List):
    """
    :type connections_to_applications: list[Detailed_Connection_To_Application]
    """

    def __init__(self, connections_to_applications):
        super().__init__(Elements.CONNECTIONS_TO_APPLICATIONS, connections_to_applications)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connections_to_applications = []
        for con_to_app_node in xml_node.iter(tag=Elements.CONNECTION_TO_APPLICATION):
            connections_to_applications.append(Detailed_Connection_To_Application.from_xml_node(con_to_app_node))
        return cls(connections_to_applications)


class ConnectionExtended(XML_Object_Base):
    def __init__(self, con_id, name, comment, open_tickets, external, status, sources, services, destinations,
                 connection_to_application):
        self.id = con_id
        self.name = name
        self.comment = comment
        self.open_tickets = open_tickets
        self.external = external
        self.status = status
        self.sources = sources
        self.services = services
        self.destinations = destinations
        self.connection_to_application = connection_to_application
        super().__init__(Elements.CONNECTION_EXTENDED)

    @classmethod
    def from_xml_node(cls, xml_node):
        con_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        open_tickets = XML_List.from_xml_node_by_tags(xml_node, Elements.OPEN_TICKETS, Elements.TICKET,
                                                      Application_Open_Ticket, True)
        external = get_xml_text_value(xml_node, Elements.EXTERNAL)
        status = get_xml_text_value(xml_node, Elements.STATUS)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        sources = []
        destinations = []
        services = []
        sources_node = get_xml_node(xml_node, Elements.SOURCES, True)
        if sources_node:
            for src_node in sources_node.iter(tag=Elements.SOURCE):
                sources.append(Network_Object.from_xml_node_auto_type(src_node))
        destinations_node = get_xml_node(xml_node, Elements.DESTINATIONS, True)
        if destinations_node:
            for dst_node in destinations_node.iter(tag=Elements.DESTINATION):
                destinations.append(Network_Object.from_xml_node_auto_type(dst_node))
        services_node = get_xml_node(xml_node, Elements.SERVICES, True)
        if services_node:
            for srv_node in services_node.iter(tag=Elements.SERVICE):
                services.append(Service_Object.from_xml_node_auto_type(srv_node))
        connection_to_application = None
        con_to_app_node = get_xml_node(xml_node, Elements.CONNECTION_TO_APPLICATION, True)
        if con_to_app_node:
            connection_to_application = Detailed_Connection_To_Application.from_xml_node(con_to_app_node)
        return cls(con_id, name, comment, open_tickets, external, status, sources, services, destinations,
                   connection_to_application)


class Detailed_Application_Connection(XML_Object_Base):
    def __init__(self, connection_id, name, external, sources, services, destinations, comment, status, open_tickets,
                 connection_to_application=None):
        self.id = connection_id
        self.name = name
        self.external = external
        self.sources = XML_List(Elements.SOURCES, sources)
        self.services = XML_List(Elements.SERVICES, services)
        self.destinations = XML_List(Elements.DESTINATIONS, destinations)
        self.comment = comment
        self.status = status
        self.open_tickets = open_tickets
        self.connection_to_application = connection_to_application
        super().__init__(Elements.CONNECTION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        sources = XML_List.from_xml_node_by_tags(xml_node, Elements.SOURCES, Elements.SOURCE, Source, True)
        services = XML_List.from_xml_node_by_tags(xml_node, Elements.SERVICES, Elements.SERVICE, Connection_Service,
                                                  True)
        destinations = XML_List.from_xml_node_by_tags(xml_node, Elements.DESTINATIONS, Elements.DESTINATION,
                                                      Destination, True)
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        status = get_xml_text_value(xml_node, Elements.STATUS)
        external = get_xml_text_value(xml_node, Elements.EXTERNAL)
        name = get_xml_text_value(xml_node, Elements.NAME)
        open_tickets = XML_List.from_xml_node_by_tags(xml_node, Elements.OPEN_TICKETS, Elements.TICKET,
                                                      Application_Open_Ticket, True)
        connection_to_application = None
        connection_to_application_node = get_xml_node(xml_node, Elements.CONNECTION_TO_APPLICATION, True)
        if connection_to_application_node:
            connection_to_application = Connection_To_Application.from_xml_node(connection_to_application_node)
        return cls(connection_id, name, external, sources, services, destinations, comment, status, open_tickets,
                   connection_to_application)

    def __repr__(self):
        srcs_string = ','.join(str(src) for src in self.sources)
        dsts_string = ','.join(str(dst) for dst in self.destinations)
        srvs_string = ','.join(str(srv) for srv in self.services)
        return 'Detailed Application Connection. Name: {}, <Sources: {}>, <Destinations: {}>, <Services: {}>'.format(
                                                                                                     self.name,
                                                                                                     srcs_string,
                                                                                                     dsts_string,
                                                                                                     srvs_string)

    def __str__(self):
        return repr(self)


class Interface_Connection(XML_Object_Base):
    def __init__(self, interface_connection_id, name, sources, services, comment, open_tickets, connected_servers):
        self.id = interface_connection_id
        self.name = name
        self.sources = sources
        self.services = services
        self.comment = comment
        self.open_tickets = open_tickets
        self.connected_servers = connected_servers
        super().__init__(Elements.INTERFACE_CONNECTION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interface_connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        sources = XML_List.from_xml_node_by_tags(xml_node, Elements.SOURCES, Elements.SOURCE, Source, True)

        services = XML_List.from_xml_node_by_tags(xml_node, Elements.SERVICES, Elements.SERVICE, Connection_Service,
                                                  True)
        connected_servers = XML_List.from_xml_node_by_tags(xml_node, Elements.CONNECTED_SERVERS, Elements.SERVER,
                                                           Server, True)
        open_tickets = XML_List.from_xml_node_by_tags(xml_node, Elements.OPEN_TICKETS, Elements.TICKET,
                                                      Application_Open_Ticket, True)

        return cls(interface_connection_id, name, sources, services, comment, open_tickets, connected_servers)


class Application_Interface(XML_Object_Base):
    def __init__(self, interface_id, name, comment, is_published, application_id, interface_connections):
        self.id = interface_id
        self.name = name
        self.comment = comment
        self.is_published = is_published
        self.application_id = application_id
        self.interface_connections = interface_connections
        super().__init__(Elements.APPLICATION_INTERFACE)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interface_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        is_published = get_xml_text_value(xml_node, Elements.IS_PUBLISHED)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        interface_connections = XML_List.from_xml_node_by_tags(xml_node, Elements.INTERFACE_CONNECTIONS,
                                                               Elements.INTERFACE_CONNECTION, Interface_Connection,
                                                               True)

        return cls(interface_id, name, comment, is_published, application_id, interface_connections)


class Application_Interfaces(XML_List):
    """
    :type application_interfaces: list[Application_Interface]
    """

    def __init__(self, application_interfaces):
        super().__init__(Elements.APPLICATION_INTERFACES, application_interfaces)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        app_interfaces = []
        for app_interface_node in xml_node.iter(tag=Elements.APPLICATION_INTERFACE):
            app_interfaces.append(Application_Interface.from_xml_node(app_interface_node))
        return cls(app_interfaces)


class Connection_Service(Base_Link_Target):
    def __init__(self, connection_id, display_name, name, link):
        super().__init__(Elements.SERVICE, connection_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(connection_id, display_name, name, link)


class Services_List(XML_List):
    """
    :type services: list[Single_Service]
    """

    def __init__(self, services):
        super().__init__(Elements.SERVICES, services)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        services = []
        for service_node in xml_node.iter(tag=Elements.SERVICE):
            if service_node.attrib[Attributes.XSI_NAMESPACE_TYPE] == Attributes.SERVICE_TYPE_SINGLE:
                services.append(Single_Service.from_xml_node(service_node))
            elif service_node.attrib[Attributes.XSI_NAMESPACE_TYPE] == Attributes.SERVICE_TYPE_GROUP:
                services.append(Group_Service.from_xml_node(service_node))
            else:
                raise ValueError(
                        "Unknown service type '{}'.".format(service_node.attrib[Attributes.XSI_NAMESPACE_TYPE]))
        return cls(services)


class Single_Service(Service_Object):
    class_identifier = Attributes.SERVICE_TYPE_SINGLE

    def __init__(self, display_name, is_global, connection_id, name, service_type, protocol, port_min, port_max, negate,
                 uid, comment, app_id=None, timeout=None):
        self.protocol = protocol
        self.min = port_min
        self.max = port_max
        self.negate = negate
        self.uid = uid
        self.comment = comment
        self.timeout = 'default' if timeout is None else timeout
        super().__init__(Elements.SERVICE, display_name, is_global, connection_id, name, service_type,
                         Attributes.SERVICE_TYPE_SINGLE, app_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        protocol = get_xml_int_value(xml_node, Elements.PROTOCOL)
        negate = get_xml_text_value(xml_node, Elements.NEGATE)
        port_min = get_xml_int_value(xml_node, Elements.MIN)
        port_max = get_xml_int_value(xml_node, Elements.MAX)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        service_type = get_xml_text_value(xml_node, Elements.TYPE)
        uid = get_xml_text_value(xml_node, Elements.UID)
        app_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        timeout = get_xml_text_value(xml_node, Elements.TIMEOUT)
        return cls(display_name, is_global, connection_id, name, service_type, protocol, port_min, port_max, negate,
                   uid, comment, app_id, timeout)

    def as_service_type(self):
        if self.protocol is not None:
            if self.min == self.max:
                return Single_Service_Type(self.protocol, self.min)
            else:
                return Range_Service_Type(self.protocol, self.min, self.max)
        else:
            return Any_Service_Type()

    def _key(self):
        return self.protocol, self.min, self.max, self.negate, self.uid, self.comment, self.timeout

    @classmethod
    def from_st_service_object(cls, st_service_object):
        return cls(st_service_object.display_name, st_service_object.global_, None,
                   st_service_object.name, st_service_object.type, st_service_object.protocol, st_service_object.min,
                   st_service_object.max, st_service_object.negate, None, st_service_object.comment)

    def __repr__(self):
        return 'type: {}, {}'.format(self.type, self.display_name)

    def __str__(self):
        return repr(self)


class Group_Service(Service_Object):
    class_identifier = Attributes.SERVICE_TYPE_GROUP

    def __init__(self, display_name, is_global, connection_id, name, service_type, members, uid, application_id=None):
        self.members = members
        self.uid = uid

        super().__init__(Elements.SERVICE, display_name, is_global, connection_id, name, service_type,
                         Attributes.SERVICE_TYPE_GROUP, application_id)

    def as_service_type(self):
        return Group_Service_Type(self.members)

    def _key(self):
        return self.id, self.uid, self.name, self.application_id, self.global_

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        service_type = get_xml_text_value(xml_node, Elements.TYPE)
        uid = get_xml_text_value(xml_node, Elements.UID)
        app_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        members = []
        for member_node in xml_node.iter(tag=Elements.MEMBER):
            member_id = get_xml_int_value(member_node, Elements.ID)
            member_display_name = get_xml_text_value(member_node, Elements.DISPLAY_NAME)
            member_name = get_xml_text_value(member_node, Elements.NAME)
            member_link = URL_Link.from_xml_node(member_node.find(Elements.LINK))
            members.append(Base_Link_Target(Elements.MEMBER, member_id, member_display_name, member_name, member_link))
        return cls(display_name, is_global, connection_id, name, service_type, members, uid, app_id)

    @classmethod
    def from_st_service_object(cls, st_service_object):
        members = []
        for member in st_service_object.members:
            members.append(Base_Link_Target(Elements.MEMBER, None, member.display_name, member.name, None))

        return cls(st_service_object.display_name, st_service_object.global_, None, st_service_object.name,
                   st_service_object.type, members, None)


    def __repr__(self):
        return 'Service Group: {}. members: {}'.format(self.display_name, ', '.join(str(member)
                                                                                    for member in self.members))

    def __str__(self):
        return repr(self)


class Source(Base_Link_Target):
    def __init__(self, connection_id, display_name, name, link, source_type):
        self.type = source_type
        super().__init__(Elements.SOURCE, connection_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        source_type = get_xml_text_value(xml_node, Elements.TYPE)
        return cls(connection_id, display_name, name, link, source_type)

    def __repr__(self):
        return self.display_name

    def __str__(self):
        return repr(self)


class Destination(Base_Link_Target):
    def __init__(self, connection_id, display_name, name, link, destination_type):
        self.type = destination_type
        super().__init__(Elements.DESTINATION, connection_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        dest_type = get_xml_text_value(xml_node, Elements.TYPE)
        return cls(connection_id, display_name, name, link, dest_type)

    def __repr__(self):
        return self.display_name

    def __str__(self):
        return repr(self)


class Network_Objects_List(XML_List):
    """
    :type network_objects: list[Network_Object]
    """

    def __init__(self, network_objects):
        super().__init__(Elements.NETWORK_OBJECTS, network_objects)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        network_objects = []
        for network_object_node in xml_node.iter(tag=Elements.NETWORK_OBJECT):
            network_objects.append(Network_Object.from_xml_node_auto_type(network_object_node))
        return cls(network_objects)


class Basic_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_BASIC

    def __init__(self, display_name, is_global, object_id, name, object_type, ip, application_id=None):

        self.ip = ip
        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         Attributes.NETWORK_OBJECT_TYPE_BASIC, application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, Elements.IP)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, object_id, name, object_type, ip, application_id)

    def as_netaddr_obj(self):
        if self.ip:
            return netaddr.IPNetwork(self.ip)
        else:
            return netaddr.IPNetwork("0.0.0.0/0")

    @classmethod
    def from_st_network_object(cls, st_network_obj):
        return cls(st_network_obj.display_name, st_network_obj.global_, None, st_network_obj.name, st_network_obj.type,
                   st_network_obj.ip, application_id=None)


class Internet_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_INTERNET

    def __init__(self, display_name, is_global, object_id, name, object_type, application_id=None):
        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         Attributes.NETWORK_OBJECT_TYPE_INTERNET, application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, object_id, name, object_type, application_id)


class Range_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_RANGE

    def __init__(self, display_name, is_global, object_id, name, object_type, first_ip, last_ip, application_id=None):
        self.first_ip = first_ip
        self.last_ip = last_ip
        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         Attributes.NETWORK_OBJECT_TYPE_RANGE, application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        first_ip = get_xml_text_value(xml_node, Elements.FIRST_IP)
        last_ip = get_xml_text_value(xml_node, Elements.LAST_IP)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, object_id, name, object_type, first_ip, last_ip, application_id)

    @classmethod
    def from_st_network_object(cls, st_network_obj):
        return cls(st_network_obj.display_name, st_network_obj.global_, None, st_network_obj.name, st_network_obj.type,
                   st_network_obj.first_ip, st_network_obj.last_ip)

    def as_netaddr_obj(self):
        return netaddr.IPRange(self.first_ip, self.last_ip)

    def __repr__(self):
        return 'Range Network Object. Name: {}, First IP: {}, Last IP: {}'.format(self.display_name, self.first_ip,
                                                                                  self.last_ip)

    def __str__(self):
        return repr(self)


class Host_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_HOST

    def __init__(self, display_name, is_global, object_id, name, object_type, ip, application_id=None):
        self.ip = ip
        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         Attributes.NETWORK_OBJECT_TYPE_HOST, application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, Elements.IP)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, object_id, name, object_type, ip, application_id)

    @classmethod
    def from_st_network_object(cls, st_network_obj):
        return cls(st_network_obj.display_name, st_network_obj.global_, None, st_network_obj.name, st_network_obj.type,
                   st_network_obj.ip)

    def as_netaddr_obj(self):
        return netaddr.IPNetwork(self.ip)

    def _key(self):
        return self.ip,

    def __repr__(self):
        return 'Host Network Object. Name: {}, IP: {}'.format(self.display_name, self.ip)

    def __str__(self):
        return repr(self)


class Subnet_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_SUBNET

    def __init__(self, display_name, is_global, object_id, name, object_type, ip, netmask, application_id=None):
        self.netmask = netmask
        self.ip = ip
        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         Attributes.NETWORK_OBJECT_TYPE_SUBNET, application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        netmask = get_xml_text_value(xml_node, Elements.NETMASK)
        ip = get_xml_text_value(xml_node, Elements.IP)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, object_id, name, object_type, ip, netmask, application_id)

    @classmethod
    def from_st_network_object(cls, st_network_obj):
        return cls(st_network_obj.display_name, st_network_obj.global_, None, st_network_obj.name, st_network_obj.type,
                   st_network_obj.ip, st_network_obj.netmask)

    def as_netaddr_obj(self):
        return netaddr.IPNetwork(str(self.ip) + "/" + str(self.netmask))

    def _key(self):
        return self.ip, self.netmask

    def __repr__(self):
        return 'Subnet Network Object. Name: {}, IP: {}, Netmask: {}'.format(self.display_name, self.ip, self.netmask)

    def __str__(self):
        return repr(self)


# TODO: Validate with real device. NOT TESTED!
class Pool_Member(XML_Object_Base):
    def __init__(self, ip, netmask, name):
        self.name = name
        self.ip = ip
        self.netmask = netmask
        super().__init__(Attributes.POOL_MEMBER)

    @classmethod
    def from_xml_node(cls, xml_node):
        name = get_xml_text_value(xml_node, Elements.NAME)
        ip = get_xml_text_value(xml_node, Elements.IP)
        netmask = get_xml_text_value(xml_node, Elements.NETMASK)
        return cls(ip, netmask, name)


class Virtual_Server_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_VIRTUAL_SERVER

    def __init__(self, display_name, is_global, object_id, name, object_type, netmask, app_id, uid, virtual_ip,
                 protocol, f5_device_name, port, comment, device_id, pool_member):
        self.netmask = netmask
        self.virtual_ip = virtual_ip
        self.netmask = netmask
        self.uid = uid
        self.protocol = protocol
        self.f5_device_name = f5_device_name
        self.port = port
        self.comment = comment
        self.device_id = device_id
        self.pool_member = pool_member
        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         Attributes.NETWORK_OBJECT_TYPE_VIRTUAL_SERVER, app_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, Elements.TYPE)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        netmask = get_xml_text_value(xml_node, Elements.NETMASK)
        virtual_ip = get_xml_text_value(xml_node, Elements.VIRTUAL_IP)
        app_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        uid = get_xml_text_value(xml_node, Elements.UID)
        protocol = get_xml_text_value(xml_node, Elements.PROTOCOL)
        f5_device_name = get_xml_text_value(xml_node, Elements.F5_DEVICE_NAME)
        port = get_xml_int_value(xml_node, Elements.PORT)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        device_id = get_xml_int_value(xml_node, Elements.DEVICE_ID)
        pool_member = XML_List.from_xml_node_by_tags(xml_node, Elements.POOL_MEMBERS, Elements.POOL_MEMBER, Pool_Member)

        return cls(display_name, is_global, object_id, name, object_type, netmask, app_id, uid, virtual_ip, protocol,
                   f5_device_name, port, comment, device_id, pool_member)

    def as_netaddr_obj(self):
        return netaddr.IPNetwork(str(self.virtual_ip) + "/" + str(self.netmask))


class VM_Instance(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_VM_INSTANCE

    def __init__(self, object_id, object_type, name, is_global, display_name,
                 instance_id, security_groups, comment, interfaces, device, application_id, ip, availability_zone,
                 vendor, tags, uid, status, host_name, original_instance_id):
        self.ip = ip
        self.instance_id = instance_id
        self.availability_zone = availability_zone
        self.vendor = vendor
        self.tags = tags
        self.uid = uid
        self.security_groups = security_groups
        self.device_info = device
        self.status = status
        self.host_name = host_name
        self.interfaces = interfaces
        self.original_instance_id = original_instance_id
        self.application_id = application_id
        self.comment = comment

        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                        Attributes.NETWORK_OBJECT_TYPE_VM_INSTANCE)

    @classmethod
    def from_xml_node(cls, xml_node):
        object_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        global_ = get_xml_text_value(xml_node, Elements.GLOBAL)
        type_ = get_xml_text_value(xml_node, Elements.TYPE)

        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, Elements.IP)
        instance_id = get_xml_text_value(xml_node, Elements.INSTANCE_ID)
        availability_zone = get_xml_text_value(xml_node, Elements.AVAILABILITY_ZONE)
        vendor = get_xml_text_value(xml_node, Elements.VENDOR)
        tags = Tags.from_xml_node(get_xml_node(xml_node, Elements.TAGS))
        uid = get_xml_text_value(xml_node, Elements.UID)
        security_groups = Security_Groups.from_xml_node(get_xml_node(xml_node, Elements.SECURITY_GROUPS))
        device_info = Device_Info.from_xml_node(get_xml_node(xml_node, Elements.DEVICE_INFO))
        status = get_xml_text_value(xml_node, Elements.STATUS)
        host_name = get_xml_text_value(xml_node, Elements.HOST_NAME)
        interfaces = Interfaces.from_xml_node(get_xml_node(xml_node, Elements.INTERFACES))
        original_instance_id = get_xml_int_value(xml_node, Elements.ORIGINAL_INSTANCE_ID)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)

        return cls(object_id, type_, name, global_, display_name, instance_id, security_groups, comment,
                   interfaces, device_info, application_id, ip, availability_zone, vendor, tags, uid, status,
                   host_name, original_instance_id)


class Tag(XML_Object_Base):
    def __init__(self, key, value):
        self.key = key
        self.value = value
        super().__init__(Elements.TAG)

    @classmethod
    def from_xml_node(cls, xml_node):
        key = get_xml_text_value(xml_node, Elements.KEY)
        value = get_xml_text_value(xml_node, Elements.VALUE)

        return cls(key, value)


class Tags(XML_List):
    def __init__(self, tags):
        super().__init__(Elements.TAGS, tags)

    @classmethod
    def from_xml_node(cls, xml_node):
        """ Initialize the object from a XML node.

        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        tags = []
        for tag_node in xml_node.iter(tag=Elements.TAG):
            tags.append(Tag.from_xml_node(tag_node))
        return cls(tags)


class VM_Instances(XML_List):
    def __init__(self, vm_instances):
        super().__init__(Elements.VM_INSTANCES, vm_instances)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        vm_instances = []
        for vm_instance_node in xml_node.iter(tag=Elements.VM_INSTANCE):
            vm_instances.append(VM_Instance.from_xml_node(vm_instance_node))
        return cls(vm_instances)


class Interface(XML_Object_Base):
    def __init__(self, name, interface_ips):
        self.name = name
        self.interface_ips = interface_ips

        super().__init__(Elements.INTERFACE)

    @classmethod
    def from_xml_node(cls, xml_node):
        name = get_xml_text_value(xml_node, Elements.NAME)
        interface_ips = Interface_IPs.from_xml_node(get_xml_node(xml_node, Elements.INTERFACE_IPS))

        return cls(name, interface_ips)


class Interfaces(XML_List):
    def __init__(self, interfaces):
        super().__init__(Elements.INTERFACES, interfaces)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interfaces = []
        for interface_node in xml_node.iter(tag=Elements.INTERFACE):
            interfaces.append(Interface.from_xml_node(interface_node))
        return cls(interfaces)


class Interface_IP(XML_Object_Base):
    def __init__(self, ip, primary, visibility):
        self.ip = ip
        self.primary = primary
        self.visibility = visibility

        super().__init__(Elements.INTERFACE_IP)

    @classmethod
    def from_xml_node(cls, xml_node):
        ip = get_xml_text_value(xml_node, Elements.IP)
        primary = get_xml_text_value(xml_node, Elements.PRIMARY)
        visibility = get_xml_text_value(xml_node, Elements.VISIBILITY)

        return cls(ip, primary, visibility)


class Interface_IPs(XML_List):
    def __init__(self, interface_ips):
        super().__init__(Elements.INTERFACE_IPS, interface_ips)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interface_ips = []
        for interface_ip_node in xml_node.iter(tag=Elements.INTERFACE_IP):
            interface_ips.append(Interface_IP.from_xml_node(interface_ip_node))
        return cls(interface_ips)


class Group_Network_Object(Network_Object):
    class_identifier = Attributes.NETWORK_OBJECT_TYPE_GROUP

    def __init__(self, display_name, is_global, connection_id, name, service_type, members, application_id=None):
        self.members = members

        super().__init__(Elements.NETWORK_OBJECT, display_name, is_global, connection_id, name, service_type,
                         Attributes.NETWORK_OBJECT_TYPE_GROUP, application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        service_type = get_xml_text_value(xml_node, Elements.TYPE)
        members = XML_List(Elements.MEMBERS, [])
        for member_node in xml_node.iter(tag=Elements.MEMBER):
            member_id = get_xml_int_value(member_node, Elements.ID)
            member_display_name = get_xml_text_value(member_node, Elements.DISPLAY_NAME)
            member_name = get_xml_text_value(member_node, Elements.NAME)
            member_link = URL_Link.from_xml_node(member_node.find(Elements.LINK))
            members.append(Base_Link_Target(Elements.MEMBER, member_id, member_display_name, member_name, member_link))
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, connection_id, name, service_type, members, application_id)

    @classmethod
    def from_st_network_object(cls, st_network_obj):
        members = []
        for member in st_network_obj.members:
            members.append(Base_Link_Target(Elements.MEMBER, None, member.display_name, member.name, None))

        sa_group_obj = cls(st_network_obj.display_name, st_network_obj.global_, None, st_network_obj.name,
                           st_network_obj.type, members)

        return sa_group_obj

    def __repr__(self):
        return 'Group Network Object. Name: {},Members: {}'.format(self.display_name,
                                                                   ','.join(str(member) for member in self.members))

    def __str__(self):
        return repr(self)


class Customers_List(XML_List):
    def __init__(self, customers):
        """
        :type customers: list[Detailed_Customer]
        """
        super().__init__(Elements.CUSTOMERS, customers)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        customers = []
        for customer_node in xml_node.iter(tag=Elements.CUSTOMER):
            customers.append(Detailed_Customer.from_xml_node(customer_node))
        return cls(customers)


class Detailed_Customer(XML_Object_Base):
    def __init__(self, customer_id, name, usage_mode, status):
        self.id = customer_id
        self.name = name
        self.usage_mode = usage_mode
        self.status = status
        super().__init__(Elements.CUSTOMER)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        customer_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        usage_mode = get_xml_text_value(xml_node, Elements.USAGE_MODE)
        status = get_xml_text_value(xml_node, Elements.STATUS)
        return cls(customer_id, name, usage_mode, status)


class Customer(Base_Link_Target):
    def __init__(self, customer_id, name, display_name=None, link=None):
        super().__init__(Elements.CUSTOMER, customer_id, display_name, name, link)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        connection_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        link = URL_Link.from_xml_node(get_xml_node(xml_node, Elements.LINK))
        return cls(connection_id, name, display_name, link)


class Device_Info(XML_Object_Base):
    def __init__(self, id_, name, region):
        self.id = id_
        self.name = name
        self.region = region
        super().__init__(Elements.DEVICE_INFO)

    @classmethod
    def from_xml_node(cls, xml_node):
        id_ = get_xml_text_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        region = get_xml_text_value(xml_node, Elements.REGION)

        return cls(id_, name, region)


class Security_Group(XML_Object_Base):
    def __init__(self, uid, name):
        self.uid = uid
        self.name = name
        super().__init__(Elements.SECURITY_GROUP)

    @classmethod
    def from_xml_node(cls, xml_node):
        uid = get_xml_text_value(xml_node, Elements.UID)
        name = get_xml_text_value(xml_node, Elements.NAME)

        return cls(uid, name)


class Security_Groups(XML_List):
    def __init__(self, security_groups):
        super().__init__(Elements.SECURITY_GROUPS, security_groups)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        security_groups = []
        for security_group_node in xml_node.iter(tag=Elements.SECURITY_GROUP):
            security_groups.append(Security_Group.from_xml_node(security_group_node))
        return cls(security_groups)




import logging

from pytos.common.base_types import XML_Object_Base, XML_List
from pytos.common.logging.definitions import XML_LOGGER_NAME
from pytos.common.definitions import xml_tags
from pytos.common.functions.xml import get_xml_text_value, get_xml_int_value, get_xml_node
from pytos.securetrack.xml_objects.rest.rules import Shadowed_Rule

logger = logging.getLogger(XML_LOGGER_NAME)


class Generic_Cleanup_List(XML_List):
    def __init__(self, count, total, score, cleanups):
        self.count = count
        self.total = total
        self.score = score
        super().__init__(xml_tags.Elements.CLEANUPS, cleanups)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        count = get_xml_int_value(xml_node, xml_tags.Elements.COUNT)
        total = get_xml_int_value(xml_node, xml_tags.Elements.TOTAL)
        score = get_xml_int_value(xml_node, xml_tags.Elements.SCORE)
        cleanups = []
        for user_node in xml_node.iter(tag=xml_tags.Elements.CLEANUP):
            cleanups.append(Generic_Cleanup.from_xml_node(user_node))
        return cls(count, total, score, cleanups)


class Generic_Cleanup(XML_Object_Base):
    def __init__(self, num_id, code, name, instances_total):
        self.id = num_id
        self.code = code
        self.name = name
        self.instances_total = instances_total
        super().__init__(xml_tags.Elements.CLEANUP)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        code = get_xml_text_value(xml_node, xml_tags.Elements.CODE)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        instances_total = get_xml_int_value(xml_node, xml_tags.Elements.INSTANCES_TOTAL)
        return cls(num_id, code, name, instances_total)


class Cleanup_Set(XML_Object_Base):
    def __init__(self, shadowed_rules_cleanup=None):
        self.shadowed_rules_cleanup = shadowed_rules_cleanup
        super().__init__(xml_tags.Elements.CLEANUP_SET)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        shadowed_rules_cleanup = Shadowed_Rules_Cleanup.from_xml_node(
                get_xml_node(xml_node, xml_tags.Elements.SHADOWED_RULES_CLEANUP))
        return cls(shadowed_rules_cleanup)


class Shadowed_Rules_Cleanup(XML_Object_Base):
    def __init__(self, shadowed_rules=None):
        self.shadowed_rules = XML_List(xml_tags.Elements.SHADOWED_RULES, shadowed_rules)
        super().__init__(xml_tags.Elements.SHADOWED_RULES_CLEANUP)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        shadowed_rules = XML_List(xml_tags.Elements.SHADOWED_RULES)
        for shadowed_rule_node in xml_node.iter(tag=xml_tags.Elements.SHADOWED_RULE):
            shadowed_rules.append(Shadowed_Rule.from_xml_node(shadowed_rule_node))
        return cls(shadowed_rules)



#!/home/user/sources/compiled/python3.4_dev/lib/python3.4

import os
import re
from setuptools import setup, find_packages

package_name = "pytos"
root_dir = os.path.abspath(os.path.dirname(__file__))


def get_version():
    file_path = os.path.join(root_dir, 'pytos', '__init__.py')
    with open(file_path, 'rt') as f:
        version = re.search('__version__ = "(\d+\.\d+\.\d+)', f.read()).group(1)
        return version


setup(
    name=package_name,
    version=get_version(),
    author="Tufin Solution Engineering",
    author_email="support@tufin.com",
    url='http://pypi.python.org/pypi/pytos',
    license="Apache License 2.0",
    description="The Tufin TOS SDK for Python",
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=[
        'netaddr>=0.7.14',
        'paramiko>=1.15.2',
        'requests>=2.6.0',
        'requests_toolbelt==0.7.1',
        'netifaces==0.10.5',
        'dnspython3==1.15.0',
        'Mako',
    ]
)



import configparser
import logging

from pytos.common.logging.definitions import COMMON_LOGGER_NAME
from pytos.common.functions.file_monitor import FileMonitor

logger = logging.getLogger(COMMON_LOGGER_NAME)


class Secure_Config_Parser(configparser.ConfigParser, FileMonitor):
    """This class is used to parse the Tufin PS library configuration files"""
    COMMON = "common"
    CUSTOM = "custom"
    SECURECHANGE = "securechange"
    SECURETRACK = "securetrack"
    LOG_LEVELS = "log_levels"
    DEFAULT_SECTIONS = (COMMON, LOG_LEVELS, SECURETRACK, SECURECHANGE)

    def __init__(self, config_file_path, custom_config_file_path=None):
        configparser.ConfigParser.__init__(self)
        self.config_file_path = config_file_path
        self.custom_config_file_path = custom_config_file_path

        if custom_config_file_path is not None:
            args = (config_file_path, custom_config_file_path)
        else:
            args = (config_file_path, )
        FileMonitor.__init__(self, args)
        self._read_config_files()

    def _read_config_files(self):
        try:
            self.read(self.config_file_path)
        except configparser.Error as config_exception:
            logger.error("Could not parse configuration file '%s'.", self.config_file_path)
            raise config_exception

        if self.custom_config_file_path is not None:
            try:
                self.read(self.custom_config_file_path)
            except configparser.Error:
                logger.error("Could not parse custom configuration file '%s'.", self.custom_config_file_path)

    def _reload_modified_file(self):
        logger.debug("Reloading modified configuration files.")
        self._read_config_files()

    def get(self, section, option, mandatory=True, raw=False, fallback=object()):
        try:
            return configparser.ConfigParser.get(self, section, option, raw=False, fallback=fallback).strip()
        except (configparser.NoOptionError, configparser.NoSectionError):
            if mandatory:
                message = "Could not find configuration option '%s' in section '%s'." % (option, section)
                logger.warning(message)
                raise KeyError(message)
            else:
                return None

    def getint(self, section, option, mandatory=True, raw=False, fallback=object()):
        try:
            return configparser.ConfigParser.getint(self, section, option, raw=False, fallback=fallback)
        except (configparser.NoOptionError, configparser.NoSectionError):
            if mandatory:
                message = "Could not find configuration option '%s' in section '%s'." % (option, section)
                logger.warning(message)
                raise KeyError(message)
            else:
                return None

    def getfloat(self, section, option, mandatory=True, raw=False, fallback=object()):
        try:
            return configparser.ConfigParser.getfloat(self, section, option, raw=False, fallback=fallback)
        except (configparser.NoOptionError, configparser.NoSectionError):
            if mandatory:
                message = "Could not find configuration option '%s' in section '%s'." % (option, section)
                logger.warning(message)
                raise KeyError(message)
            else:
                return None

    def getboolean(self, section, option, mandatory=True, raw=False, fallback=object()):
        try:
            return configparser.ConfigParser.getboolean(self, section, option, raw=False, fallback=fallback)
        except (configparser.NoOptionError, configparser.NoSectionError):
            if mandatory:
                message = "Could not find configuration option '%s' in section '%s'." % (option, section)
                logger.warning(message)
                raise KeyError(message)
            else:
                return None

    def dict(self, section):
        return dict(configparser.ConfigParser.items(self, section))

    def update_config_file(self):
        """
        Write the configuration values from memory to the configuration file.
        """
        logger.info("Updating configuration file '%s' with configuration.", self.config_file_path)
        with open(self.config_file_path, "w") as config_file:
            delimiter = " {} ".format(self._delimiters[0])
            if self._defaults:
                self._write_section(config_file, self.default_section, self._defaults.items(), delimiter)
            for section in self._sections:
                if section in Secure_Config_Parser.DEFAULT_SECTIONS:
                    self._write_section(config_file, section, self._sections[section].items(), delimiter)



from pytos.securechange.xml_objects.restapi.step.initialize import *

logger = logging.getLogger(XML_LOGGER_NAME)


class AbsNetwork(XML_Object_Base, metaclass=SubclassWithIdentifierRegistry):
    """Base class for parsing all network object"""

    @classmethod
    def from_xml_node(cls, xml_node):
        try:
            network_type = xml_node.attrib[Attributes.XSI_NAMESPACE_TYPE]
        except KeyError:
            msg = 'XML node is missing the XSI attribute "{}"'.format(Attributes.XSI_NAMESPACE_TYPE)
            logger.error(msg)
            raise ValueError(msg)
        else:
            try:
                return cls.registry[network_type](xml_node)
            except KeyError:
                logger.error('Unknown violation object type "{}"'.format(network_type))


class NetworkObject(AbsNetwork):
    """Base class for all sub type of the network object"""

    def __init__(self, xml_node, element):
        self.address_book = get_xml_text_value(xml_node, Elements.ADDRESS)
        self.type_on_device = get_xml_text_value(xml_node, Elements.TYPE)
        self.version_id = get_xml_int_value(xml_node, Elements.VERSION_ID)
        self.referenced = get_xml_text_value(xml_node, Elements.REFERENCED)
        interface_name = get_xml_text_value(xml_node, Elements.INTERFACE_NAME)
        self.nat_info = NatInfo(interface_name)
        self.installable_target = get_xml_text_value(xml_node, Elements.INSTALLABLE_TARGET)
        self.group_id = get_xml_text_value(xml_node, Elements.GROUP_ID)
        self.device_type = get_xml_text_value(xml_node, Elements.DEVICE_TYPE)
        self.ip_type = get_xml_text_value(xml_node, Elements.IP_TYPE)
        self.id = get_xml_text_value(xml_node, Elements.ID)
        zone_node = get_xml_node(xml_node, Elements.ZONE, True)
        if zone_node is not None:
            self.zone = PolicyZone(zone_node)
        else:
            self.zone = None
        self.device_id = get_xml_int_value(xml_node, Elements.DEVICE_ID)
        admin_domain_node = get_xml_node(xml_node, Elements.ADMIN_DOMAIN, True)
        if admin_domain_node is not None:
            self.admin_domain = AdminDomain.from_xml_node(admin_domain_node)
        else:
            self.admin_domain = None
        self.inDomainElementId = get_xml_text_value(xml_node, Elements.INDOMAINELEMENTID)
        self.global_el = Flat_XML_Object_Base(Elements.GLOBAL, None, get_xml_text_value(xml_node, Elements.GLOBAL))
        self.origin = get_xml_text_value(xml_node, Elements.ORIGIN)
        self.comment = get_xml_text_value(xml_node, Elements.COMMENT)
        self.shared = get_xml_text_value(xml_node, Elements.SHARED)
        self.name = get_xml_text_value(xml_node, Elements.NAME)
        self.implicit = get_xml_text_value(xml_node, Elements.IMPLICIT)
        self.class_name = get_xml_text_value(xml_node, Elements.CLASS_NAME)
        self.display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        self.uid = get_xml_text_value(xml_node, Elements.UID)
        self.any_zone = get_xml_text_value(xml_node, Elements.ANY_ZONE)
        self.management_domain = get_xml_text_value(xml_node, Elements.MANAGEMENT_DOMAIN)
        self.domain_id = get_xml_int_value(xml_node, Elements.DOMAIN_ID)
        self.application_name = get_xml_text_value(xml_node, Elements.APPLICATION_NAME)
        super().__init__(element)

    def __str__(self):
        return self.display_name


class AnyNetworkObject(NetworkObject):
    """The class represents the any_network_object"""
    class_identifier = Attributes.VIOLATION_ANY_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.VIOLATION_ANY_NETWORK_OBJECT)

    def __str__(self):
        return "Any"


class HostNetworkObject(NetworkObject):
    """The class represents the host_network_object"""
    class_identifier = Attributes.HOST_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.ip = get_xml_text_value(xml_node, Elements.IP)
        self.subnet_mask = get_xml_text_value(xml_node, Elements.SUBNET_MASK)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.HOST_NETWORK_OBJECT)

    def __str__(self):
        return self.ip


class SubnetNetworkObject(NetworkObject):
    """The class represents the subnet_network_object"""
    class_identifier = Attributes.SUBNET_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.ip = get_xml_text_value(xml_node, Elements.IP)
        self.subnet_mask = get_xml_text_value(xml_node, Elements.SUBNET_MASK)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.SUBNET_NETWORK_OBJECT)

    def __str__(self):
        return "{}/{}".format(self.ip, self.subnet_mask)


class RangeNetworkObject(NetworkObject):
    class_identifier = Attributes.RANGE_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.RANGE_NETWORK_OBJECT)
        self.min_ip = get_xml_text_value(xml_node, Elements.MIN_IP)
        self.max_ip = get_xml_text_value(xml_node, Elements.MAX_IP)

    def __str__(self):
        return self.min_ip + '-' + self.max_ip


class NetworkObjectGroup(NetworkObject):
    """The class represents the subnet_network_object"""
    class_identifier = Attributes.NETWORK_OBJECT_GROUP

    def __init__(self, xml_node):
        self.members = []
        for member_node in xml_node.iter(tag=Elements.MEMBER):
            self.members.append(NetworkObject.from_xml_node(member_node))

        self.exclusions = []
        for member_node in xml_node.iter(tag=Elements.EXCLUSION):
            self.exclusions.append(NetworkObject.from_xml_node(member_node))

        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.NETWORK_OBJECT_GROUP)


class DomainNetworkObject(NetworkObject):
    class_identifier = Attributes.DOMAIN_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.DOMAIN_NETWORK_OBJECT)


class InstallOnNetworkObject(NetworkObject):
    class_identifier = Attributes.INSTALL_ON_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.INSTALL_ON_NETWORK_OBJECT)


class HostNetworkObjectWithInterfaces(NetworkObject):
    class_identifier = Attributes.HOST_NETWORK_OBJECT_WITH_INTERFACES

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.HOST_NETWORK_OBJECT_WITH_INTERFACES)
        self.ip = get_xml_text_value(xml_node, Elements.IP)
        self.subnet_mask = get_xml_text_value(xml_node, Elements.SUBNET_MASK)
        self.interfaces = []
        for member_node in xml_node.iter(tag=Elements.INTERFACE_FOR_NETWORK_OBJECT):
            self.interfaces.append(NetworkObject.from_xml_node(member_node))


class CloudSecurityGroup(NetworkObject):
    class_identifier = Attributes.CLOUD_SECURITY_GROUP_NETWORK_OBJECT

    def __init__(self, xml_node):
        self.members = []
        for member_node in xml_node.iter(tag=Elements.MEMBER):
            self.members.append(NetworkObject.from_xml_node(member_node))

        self.exclusions = []
        for member_node in xml_node.iter(tag=Elements.EXCLUSION):
            self.exclusions.append(NetworkObject.from_xml_node(member_node))

        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.CLOUD_SECURITY_GROUP_NETWORK_OBJECT)


class InternetNetworkObject(NetworkObject):
    class_identifier = Attributes.INTERNET_NETWORK_OBJECT

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.INTERNET_NETWORK_OBJECT)


class AbsService(XML_Object_Base, metaclass=SubclassWithIdentifierRegistry):
    """Base class for parsing all services objects"""

    @classmethod
    def from_xml_node(cls, xml_node):
        if xml_node is None:
            return None
        try:
            service_type = xml_node.attrib[Attributes.XSI_NAMESPACE_TYPE]
        except KeyError:
            msg = 'XML node is missing the XSI attribute "{}"'.format(Attributes.XSI_NAMESPACE_TYPE)
            logger.error(msg)
            raise ValueError(msg)
        else:
            try:
                return cls.registry[service_type](xml_node)
            except KeyError:
                logger.error('Unknown violation object type "{}"'.format(service_type))


class Service(AbsService):
    """Base class for all sub type of the services objects"""

    def __init__(self, xml_node, element):
        self.version_id = get_xml_text_value(xml_node, Elements.VERSION_ID)
        self.referenced = get_xml_text_value(xml_node, Elements.REFERENCED)
        self.match_rule = get_xml_text_value(xml_node, Elements.MATCH_RULE)
        self.id = get_xml_text_value(xml_node, Elements.ID)
        self.device_id = get_xml_int_value(xml_node, Elements.DEVICE_ID)
        self.admin_domain = AdminDomain.from_xml_node(xml_node)
        self.in_domain_element_id = get_xml_text_value(xml_node, Elements.INDOMAINELEMENTID)
        self.global_el = Flat_XML_Object_Base(Elements.GLOBAL, None, get_xml_text_value(xml_node, Elements.GLOBAL))
        self.origin = get_xml_text_value(xml_node, Elements.ORIGIN)
        self.comment = get_xml_text_value(xml_node, Elements.COMMENT)
        self.shared = get_xml_text_value(xml_node, Elements.SHARED)
        self.name = get_xml_text_value(xml_node, Elements.NAME)
        self.implicit = get_xml_text_value(xml_node, Elements.IMPLICIT)
        self.class_name = get_xml_text_value(xml_node, Elements.CLASS_NAME)
        self.display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        self.uid = get_xml_text_value(xml_node, Elements.UID)
        super().__init__(element)

    def __str__(self):
        return self.display_name


class AnyService(Service):
    """The class represents the any_service_object"""
    class_identifier = Attributes.VIOLATION_ANY_SERVICE

    def __init__(self, xml_node):
        self.negate = get_xml_text_value(xml_node, Elements.NEGATE)
        self.match_for_any = get_xml_text_value(xml_node, Elements.MATCH_FOR_ANY)
        self.timeout = get_xml_text_value(xml_node, Elements.TIMEOUT)
        self.min_protocol = get_xml_int_value(xml_node, Elements.MIN_PROTOCOL)
        self.max_protocol = get_xml_int_value(xml_node, Elements.MAX_PROTOCOL)
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.VIOLATION_ANY_SERVICE)


class TransportService(Service):
    """The class represents the transport_service_object"""
    class_identifier = Attributes.TRANSPORT_SERVICE

    def __init__(self, xml_node):
        self.cp_inspect_streaming_name = get_xml_text_value(xml_node, Elements.CP_INSPECT_STREAMING_NAME)
        self.min_protocol = get_xml_int_value(xml_node, Elements.MIN_PROTOCOL)
        self.max_protocol = get_xml_int_value(xml_node, Elements.MAX_PROTOCOL)
        self.min_port = get_xml_int_value(xml_node, Elements.MIN_PORT)
        self.max_port = get_xml_int_value(xml_node, Elements.MAX_PORT)
        self.protocol = get_xml_int_value(xml_node, Elements.PROTOCOL)
        self.min_value_source = get_xml_int_value(xml_node, Elements.MIN_VALUE_SOURCE)
        self.max_value_source = get_xml_int_value(xml_node, Elements.MAX_VALUE_SOURCE)
        self.cp_prototype_name = get_xml_text_value(xml_node, Elements.CP_PROTOTYPE_NAME)
        self.match_for_any = get_xml_text_value(xml_node, Elements.MATCH_FOR_ANY)
        self.negate = get_xml_text_value(xml_node, Elements.NEGATE)
        self.timeout = get_xml_text_value(xml_node, Elements.TIMEOUT)
        self.display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.TRANSPORT_SERVICE)


class IcmpService(Service):
    """The class represents the icmp_service_object"""
    class_identifier = Attributes.ICMP_SERVICE

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.ICMP_SERVICE)
        self.type_on_device = get_xml_text_value(xml_node, Elements.TYPE_ON_DEVICE)
        self.negate = get_xml_text_value(xml_node, Elements.NEGATE)
        self.match_for_any = get_xml_text_value(xml_node, Elements.MATCH_FOR_ANY)
        self.timeout = get_xml_int_value(xml_node, Elements.TIMEOUT)
        self.min_icmp_type = get_xml_int_value(xml_node, Elements.MIN_ICMP_TYPE)
        self.max_icmp_type = get_xml_int_value(xml_node, Elements.MAX_ICMP_TYPE)


class IPService(Service):
    """The class represents the ip_service_object"""
    class_identifier = Attributes.IP_SERVICE

    def __init__(self, xml_node):
        self.negate = get_xml_text_value(xml_node, Elements.NEGATE)
        self.match_for_any = get_xml_text_value(xml_node, Elements.MATCH_FOR_ANY)
        self.timeout = get_xml_text_value(xml_node, Elements.TIMEOUT)
        self.min_protocol = get_xml_int_value(xml_node, Elements.MIN_PROTOCOL)
        self.max_protocol = get_xml_int_value(xml_node, Elements.MAX_PROTOCOL)
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.IP_SERVICE)


class ServiceGroup(Service):
    """The class represents the ip_service_object"""
    class_identifier = Attributes.SERVICE_GROUP

    def __init__(self, xml_node):
        super().__init__(xml_node, xml_node.find('.').tag)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.SERVICE_GROUP)
        self.members = [Service.from_xml_node(node) for node in xml_node.findall('member')]


class Binding(XML_Object_Base, metaclass=SubclassWithIdentifierRegistry):
    """Base Binding Class that handles all Binding sub Binding DTO parsing"""

    @classmethod
    def from_xml_node(cls, xml_node):
        if xml_node is None:
            return None
        try:
            binding_type = xml_node.attrib[Attributes.XSI_NAMESPACE_TYPE]
        except KeyError:
            msg = 'XML node is missing the XSI attribute "{}"'.format(Attributes.XSI_NAMESPACE_TYPE)
            logger.error(msg)
            raise ValueError(msg)
        else:
            try:
                return cls.registry[binding_type](xml_node)
            except KeyError:
                logger.error('Unknown binding object type "{}"'.format(binding_type))


class AclBinding(Binding):
    """The class represents the acl_binding_object which is sub type of Binding_DTO"""
    class_identifier = Attributes.ACL__BINDING

    def __init__(self, xml_node):
        self.acl_name = get_xml_text_value(xml_node, Elements.ACL_NAME)

        self.incoming_interface_names = [node.text for node in xml_node.iter(Elements.INCOMING_INTERFACE_NAME)]
        self.outgoing_interface_names = [node.text for node in xml_node.iter(Elements.OUTGOING_INTERFACE_NAME)]

        super().__init__(Elements.BINDING)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.ACL__BINDING)


class ZoneBinding(Binding):
    """The class represents the zone_binding object which is sub type of Binding_DTO"""
    class_identifier = Attributes.ZONE__BINDING

    def __init__(self, xml_node):
        self.from_zone = get_xml_text_value(xml_node, Elements.FROM_ZONE)
        self.to_zone = get_xml_text_value(xml_node, Elements.TO_ZONE)
        super().__init__(Elements.BINDING)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.ZONE__BINDING)


class PolicyBinding(Binding):
    class_identifier = Attributes.POLICY__BINDING

    def __init__(self, xml_node):
        self.policy_name = get_xml_text_value(xml_node, Elements.POLICY_NAME)
        self.installed_on_module = get_xml_text_value(xml_node, Elements.INSTALLED_ON_MODULE)
        super().__init__(Elements.BINDING)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.POLICY__BINDING)


class AbsSlimRule(XML_Object_Base, metaclass=SubclassWithIdentifierRegistry):
    """AbsSlimRule Class that handles all SlimRule sub DTO parsing"""
    @classmethod
    def from_xml_node(cls, xml_node):
        if xml_node is None:
            return None
        try:
            rule_type = xml_node.attrib[Attributes.XSI_NAMESPACE_TYPE]
        except KeyError:
            msg = 'XML node is missing the XSI attribute "{}"'.format(Attributes.XSI_NAMESPACE_TYPE)
            logger.error(msg)
            raise ValueError(msg)
        else:
            try:
                return cls.registry[rule_type].from_xml_node(xml_node)
            except KeyError:
                logger.error('Unknown binding object type "{}"'.format(rule_type))


class SlimRule(AbsSlimRule):
    """The class represents the SlimRule which is sub type of SlimRule"""
    def __init__(self, uid, destination_networks=None, source_networks=None,
                 destination_services=None, rule_number=None, additional_parameters=None, communities=None,
                 rule_location=None, applications=None, install_ons=None, users=None, track=None, source_services=None,
                 from_zone=None, to_zone=None, action=None, comment=None, name=None, is_disabled=None):
        self.additional_parameters = additional_parameters
        self.communities = communities
        self.sourceNetworks = source_networks
        self.destinationNetworks = destination_networks
        self.destination_services = destination_services
        self.install_ons = install_ons
        self.track = track
        self.rule_location = rule_location
        self.source_services = source_services
        self.uid = uid
        self.rule_number = rule_number
        if applications is not None:
            self.applications = applications
        self.from_zone = from_zone
        self.to_zone = to_zone
        self.action = action
        self.comment = comment
        self.name = name
        self.is_disabled = is_disabled
        self.users = users
        super().__init__(Elements.RULE)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.SLIM_RULE_WITH_META_DATA)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        rule_uid = get_xml_text_value(xml_node, Elements.UID)
        rule_location = get_xml_text_value(xml_node, Elements.RULE_LOCATION)
        rule_number = get_xml_int_value(xml_node, Elements.RULENUMBER)
        from_zone = get_xml_text_value(xml_node, Elements.FROM_ZONE)
        to_zone = get_xml_text_value(xml_node, Elements.TO_ZONE)
        action = get_xml_text_value(xml_node, Elements.ACTION)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        name = get_xml_text_value(xml_node, Elements.NAME)
        is_disabled = get_xml_text_value(xml_node, Elements.ISDISABLED)

        destination_networks = []
        for destination_network_node in xml_node.iter(tag=Elements.DESTNETWORKS):
            network_object = NetworkObject.from_xml_node(destination_network_node)
            destination_networks.append(network_object)

        source_networks = []
        for source_network_node in xml_node.iter(tag=Elements.SOURCENETWORKS):
            network_object = NetworkObject.from_xml_node(source_network_node)
            source_networks.append(network_object)

        destination_services = []
        for destination_service_node in xml_node.iter(Elements.DESTINATIONSERVICES):
            service = Service.from_xml_node(destination_service_node)
            destination_services.append(service)

        additional_parameters = []
        parameters_node = get_xml_node(xml_node, Elements.ADDITIONAL_PARAMETERS, True)
        if parameters_node is not None:
            for parameter_node in parameters_node.iter(Elements.ADDITIONAL_PARAMETER):
                additional_parameter = AdditionalParameter.from_xml_node(parameter_node)
                additional_parameters.append(additional_parameter)

        communities = []
        communities_node = get_xml_node(xml_node, Elements.COMMUNITIES, True)
        if communities_node is not None:
            for community_node in communities_node.iter(Elements.COMMUNITY):
                community = VpnCommunity.from_xml_node(community_node)
                communities.append(community)

        applications = []
        applications_node = get_xml_node(xml_node, Elements.APPLICATIONS, True)
        if applications_node is not None:
            for application_node in applications_node.iter(Elements.APPLICATION):
                application = Application.from_xml_node(application_node)
                applications.append(application)

        install_ons = []
        install_ons_node = get_xml_node(xml_node, Elements.INSTALL_ONS, True)
        if install_ons_node is not None:
            for install_on_node in install_ons_node.iter(tag=Elements.INSTALL_ON):
                network_object = NetworkObject.from_xml_node(install_on_node)
                install_ons.append(network_object)

        users = []
        users_node = get_xml_node(xml_node, Elements.USERS, True)
        if users_node is not None:
            for user_node in users_node.iter(tag=Elements.USER):
                user = DeviceUser.from_xml_node(user_node)
                users.append(user)

        track_node = get_xml_node(xml_node, Elements.TRACK, True)
        if track_node is not None:
            track = RuleTrack.from_xml_node(track_node)
        else:
            track = None

        source_services = []
        for source_service_node in xml_node.iter(Elements.SOURCESERVICES):
            service = Service.from_xml_node(source_service_node)
            source_services.append(service)

        return cls(rule_uid, destination_networks, source_networks, destination_services,
                   rule_number, additional_parameters, communities, rule_location, applications, install_ons,
                   users, track, source_services, from_zone, to_zone, action, comment, name, is_disabled)

    def to_pretty_str(self):
        rule_string = "Rule name: {}\n".format(self.name)
        rule_string += "From zone: {}\n".format(self.from_zone)
        rule_string += "To zone: {}\n".format(self.to_zone)
        rule_string += "Sources: {}\n".format(", ".join(str(src) for src in self.sourceNetworks))
        rule_string += "Destinations: {}\n".format(", ".join(str(src) for src in self.destinationNetworks))
        rule_string += "Services: {}\n".format(", ".join(str(srv) for srv in self.destination_services))
        if self.comment is not None:
            rule_string += "Comment: {}\n".format(unescape(self.comment))
        return rule_string


class SlimRuleWithMetadata(SlimRule):
    """This class represents the SlimRuleWithMetadata"""

    def __init__(self, uid, destination_networks=None, source_networks=None,
                 destination_services=None, rule_number=None, additional_parameters=None, communities=None,
                 rule_location=None, applications=None, install_ons=None, users=None, track=None, source_services=None,
                 from_zone=None, to_zone=None, action=None, comment=None, name=None, is_disabled=None, rule_meta_data=None):
        super().__init__(uid, destination_networks, source_networks,
                 destination_services, rule_number, additional_parameters, communities,
                 rule_location, applications, install_ons, users, track, source_services,
                 from_zone, to_zone, action, comment, name, is_disabled)
        self.rule_meta_data = rule_meta_data
        self.set_attrib(Attributes.XSI_TYPE, Attributes.SLIM_RULE_WITH_META_DATA)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        slim_rule = super().from_xml_node(xml_node)
        rule_meta_data_node = get_xml_node(xml_node, Elements.RULE_METADATA, True)
        if rule_meta_data_node is not None:
            slim_rule.rule_meta_data = RuleMetaData.from_xml_node(rule_meta_data_node)
        else:
            slim_rule.rule_meta_data = None
        return slim_rule

    def to_pretty_str(self):
        rule_string = super().to_pretty_str()
        rule_string += self.rule_meta_data.to_pretty_str()
        return rule_string


class RuleMetaData(XML_Object_Base):
    """This class represents the RuleMetaData used in rule decommission field"""

    def __init__(self, violations, permissiveness_level, legacy_rule, ticket_ids, tech_owner, last_hit,
                 rule_description, business_owners, last_modified, shadowed_status, applications):
        self.violations = violations
        self.permissiveness_level = permissiveness_level
        self.legacy_rule = legacy_rule
        self.ticket_ids = ticket_ids
        self.tech_owner = tech_owner
        self.last_hit = last_hit
        self.rule_description = rule_description
        self.business_owners = business_owners
        self.last_modified = last_modified
        self.shadowed_status = shadowed_status
        self.applications = applications
        super().__init__(Elements.RULE_METADATA)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        violations = get_xml_text_value(xml_node, Elements.VIOLATIONS)
        permissiveness_level = get_xml_text_value(xml_node, Elements.PERMISSIVENESS_LEVEL)
        legacy_rule = get_xml_text_value(xml_node, Elements.LEGACY_RULE)
        ticket_ids = get_xml_text_value(xml_node, Elements.TICKET_IDS)
        last_hit = get_xml_text_value(xml_node, Elements.LAST_HIT)
        rule_description = get_xml_text_value(xml_node, Elements.RULE_DESCRIPTION)
        business_owners = get_xml_text_value(xml_node, Elements.BUSINESS_OWNERS)
        last_modified = get_xml_text_value(xml_node, Elements.LAST_MODIFIED)
        shadowed_status = get_xml_text_value(xml_node, Elements.SHADOWED_STATUS)
        tech_owner = get_xml_text_value(xml_node, Elements.TECH_OWNER)
        applications = []
        applications_node = get_xml_node(xml_node, Elements.APPLICATIONS, True)
        if applications_node is not None:
            for application_node in applications_node.iter(Elements.APPLICATION):
                application = SaApplication.from_xml_node(application_node)
                applications.append(application)

        return cls(violations, permissiveness_level, legacy_rule, ticket_ids, tech_owner, last_hit, rule_description,
                   business_owners, last_modified, shadowed_status, applications)

    def to_pretty_str(self):
        meta_data_string = ''
        if self.violations is not None:
            meta_data_string += "Violations: {}\n".format(self.violations)
        if self.permissiveness_level is not None:
            meta_data_string += "Permissiveness level: {}\n".format(self.permissiveness_level)
        if self.legacy_rule is not None:
            meta_data_string += "Legacy rule: {}\n".format(self.legacy_rule)
        if self.ticket_ids is not None:
            meta_data_string += "Ticket IDs: {}\n".format(self.ticket_ids)
        if self.tech_owner is not None:
            meta_data_string += "Tech Owner: {}\n".format(self.tech_owner)
        if self.last_hit is not None:
            meta_data_string += "Last hit: {}\n".format(self.last_hit)
        if self.rule_description is not None:
            meta_data_string += "Rule description: {}\n".format(self.rule_description)
        if self.business_owners is not None:
            meta_data_string += "Business owners: {}\n".format(self.business_owners)
        if self.last_modified is not None:
            meta_data_string += "Last modified: {}\n".format(self.last_modified)
        if self.shadowed_status is not None:
            meta_data_string += "Shadowed status: {}\n".format(self.shadowed_status)
        if self.applications:
            meta_data_string += "Applications: {}\n".format(", ".join(str(app) for app in self.applications))
        return meta_data_string


class AdditionalParameter(XML_Object_Base):
    """This class represents the Additional Parameter used in rule decommission field"""

    def __init__(self, num_id, display_name, class_name, name, uid):
        self.id = num_id
        self.display_name = display_name
        self.class_name = class_name
        self.name = name
        self.uid = uid
        super().__init__(Elements.ADDITIONAL_PARAMETER)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        class_name = get_xml_text_value(xml_node, Elements.CLASS_NAME)
        name = get_xml_text_value(xml_node, Elements.NAME)
        uid = get_xml_text_value(xml_node, Elements.UID)
        return cls(num_id, display_name, class_name, name, uid)


class VpnCommunity(XML_Object_Base):
    """This class represents the VpnCommunity used in rule decommission field"""

    def __init__(self, class_name, name, uid):
        self.class_name = class_name
        self.name = name
        self.uid = uid
        super().__init__(Elements.COMMUNITIES)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        class_name = get_xml_text_value(xml_node, Elements.CLASS_NAME)
        name = get_xml_text_value(xml_node, Elements.NAME)
        uid = get_xml_text_value(xml_node, Elements.UID)
        return cls(class_name, name, uid)


class Application(XML_Object_Base):
    """This class represents the Application used in rule decommission field"""

    def __init__(self, application_name, in_domain_element_id, domain_id, device_id, admin_domain, a_global, origin,
                 comment, shared, name, implicit, class_name, display_name, uid):
        self.application_name = application_name
        self.inDomainElementId = in_domain_element_id
        self.domain_id = domain_id
        self.device_id = device_id
        self.admin_domain = admin_domain
        self.a_global = a_global
        self.origin = origin
        self.comment = comment
        self.shared = shared
        self.name = name
        self.implicit = implicit
        self.class_name = class_name
        self.display_name = display_name
        self.uid = uid
        super().__init__(Elements.APPLICATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        application_name = get_xml_text_value(xml_node, Elements.APPLICATION_NAME)
        in_domain_element_id = get_xml_text_value(xml_node, Elements.INDOMAINELEMENTID)
        domain_id = get_xml_int_value(xml_node, Elements.DOMAIN_ID)
        device_id = get_xml_int_value(xml_node, Elements.DEVICE_ID)
        admin_domain_node = get_xml_node(xml_node, Elements.ADMIN_DOMAIN, True)
        if admin_domain_node is not None:
            admin_domain = AdminDomain.from_xml_node(admin_domain_node)
        else:
            admin_domain = None
        a_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        origin = get_xml_text_value(xml_node, Elements.ORIGIN)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        shared = get_xml_text_value(xml_node, Elements.SHARED)
        name = get_xml_text_value(xml_node, Elements.NAME)
        implicit = get_xml_text_value(xml_node, Elements.IMPLICIT)
        class_name = get_xml_text_value(xml_node, Elements.CLASS_NAME)
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        uid = get_xml_text_value(xml_node, Elements.UID)

        return cls(application_name, in_domain_element_id, domain_id, device_id, admin_domain, a_global, origin,
                   comment, shared, name, implicit, class_name, display_name, uid)


class AdminDomain(XML_Object_Base):
    """This class represents the AdminDomain used in rule decommission field"""

    def __init__(self, name, uid):
        self.name = name
        self.uid = uid
        super().__init__(Elements.ADMIN_DOMAIN)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, Elements.NAME)
        uid = get_xml_text_value(xml_node, Elements.UID)
        return cls(name, uid)


class NatInfo(XML_Object_Base):
    def __init__(self, interface_name):
        self.interface_name = interface_name
        super().__init__(Elements.NAT_INFO)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interface_name = get_xml_text_value(xml_node, Elements.INTERFACE_NAME)
        return cls(interface_name)


class PolicyZone(XML_Object_Base):
    """The class represents the PolicyZoneDTO"""
    def __init__(self, xml_node):
        self.zone_name_in_parent = get_xml_text_value(xml_node, Elements.ZONE_NAME_IN_PARENT)
        self.address_book = get_xml_text_value(xml_node, Elements.ADDRESS_BOOK)
        self.version_id = get_xml_int_value(xml_node, Elements.VERSION_ID)
        self.admin_domain = AdminDomain.from_xml_node(xml_node)
        self.global_el = Flat_XML_Object_Base(Elements.GLOBAL, None, get_xml_text_value(xml_node, Elements.GLOBAL))
        self.name = get_xml_text_value(xml_node, Elements.NAME)
        super().__init__(Elements.ZONE)


class DeviceUser(XML_Object_Base):
    """This class represents the DeviceUser used in rule decommission field"""

    def __init__(self, class_name, name, uid):
        self.class_name = class_name
        self.name = name
        self.uid = uid
        super().__init__(Elements.USERS)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        class_name = get_xml_text_value(xml_node, Elements.CLASS_NAME)
        name = get_xml_text_value(xml_node, Elements.NAME)
        uid = get_xml_text_value(xml_node, Elements.UID)
        return cls(class_name, name, uid)


class RuleTrack(XML_Object_Base):
    """This class represents the RuleTrack used in rule decommission field"""

    def __init__(self, track_interval, track_level):
        self.track_interval = track_interval
        self.track_level = track_level
        super().__init__(Elements.TRACK)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        track_interval = get_xml_text_value(xml_node, Elements.TRACK_INTERVAL)
        track_level = get_xml_text_value(xml_node, Elements.TRACK_LEVEL)
        return cls(track_interval, track_level)


class SaApplication(XML_Object_Base):
    """This class represents the aApplication used in rule decommission field"""

    def __init__(self, num_id, domain_id, name, owner):
        self.id = num_id
        self.domain_id = domain_id
        self.name = name
        self.owner = owner
        super().__init__(Elements.APPLICATIONS)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        domain_id = get_xml_int_value(xml_node, Elements.DOMAIN_ID)
        owner = get_xml_text_value(xml_node, Elements.OWNER)
        return cls(num_id, name, domain_id, owner)

    def __str__(self):
        return self.name




import logging
import xml.etree.ElementTree as ET
import netaddr

from pytos.common.base_types import XML_Object_Base, SubclassWithIdentifierRegistry, Comparable
from pytos.common.definitions.xml_tags import Attributes, Elements, XSI_NAMESPACE_URL, NAMESPACE_FIELD_ATTRIB_CONTENT
from pytos.common.functions import str_to_bool, XML_LOGGER_NAME
from pytos.common.functions.xml import get_xml_text_value, get_xml_int_value

logger = logging.getLogger(XML_LOGGER_NAME)


class Base_Object(XML_Object_Base, Comparable):
    def __init__(self, xml_tag, display_name, is_global, object_id, name, object_type, application_id=None):
        self.id = object_id
        self.name = name
        self.type = object_type
        self.display_name = display_name
        self.global_ = is_global
        self.application_id = application_id
        super().__init__(xml_tag)
        self.set_attrib(NAMESPACE_FIELD_ATTRIB_CONTENT, XSI_NAMESPACE_URL)


    def is_global(self):
        return str_to_bool(self.global_)

    def _key(self):
        return self.id, self.name, self.type, self.global_, self.application_id


class Base_Link_Target(XML_Object_Base, Comparable):
    def __init__(self, xml_tag, connection_id, display_name, name, link):
        self.id = connection_id
        self.display_name = display_name
        self.name = name
        self.link = link
        super().__init__(xml_tag)

    def get_reference_link(self):
        return self.link.get_attrib(Attributes.HREF)

    def _key(self):
        return self.name, self.id

    def __str__(self):
        return self.display_name


class Base_Service(Base_Object):
    def __init__(self, display_name, is_global, service_id, name, service_type, uid, application_id=None):
        self.uid = uid
        super().__init__(Elements.SERVICE, display_name, is_global, service_id, name, service_type,
                         application_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        display_name = get_xml_text_value(xml_node, Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, Elements.GLOBAL)
        service_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        service_type = get_xml_text_value(xml_node, Elements.TYPE)
        uid = get_xml_text_value(xml_node, Elements.UID)
        application_id = get_xml_int_value(xml_node, Elements.APPLICATION_ID)
        return cls(display_name, is_global, service_id, name, service_type, uid, application_id)


class URL_Link(XML_Object_Base):
    def __init__(self, url):
        super().__init__(Elements.LINK)
        self.set_attrib(NAMESPACE_FIELD_ATTRIB_CONTENT, XSI_NAMESPACE_URL)
        self.set_attrib(Attributes.HREF, url)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        url = xml_node.attrib[Attributes.HREF]
        return cls(url)


class Network_Object(Base_Object, metaclass=SubclassWithIdentifierRegistry):
    def __init__(self, xml_tag, display_name, is_global, object_id, name, object_type, attr_type, application_id=None):
        super().__init__(xml_tag, display_name, is_global, object_id, name, object_type, application_id)
        self.set_attrib(Attributes.XSI_TYPE, attr_type)

    def as_netaddr_obj(self):
        raise NotImplementedError

    def as_netaddr_set(self):
        """This returns a netaddr set representing the Network_Object"""
        return netaddr.IPSet(self.as_netaddr_obj())


    @staticmethod
    def from_xml_string_auto_type(xml_string):
        xml_node = ET.fromstring(xml_string)
        return Network_Object.from_xml_node_auto_type(xml_node)

    @classmethod
    def from_xml_node_auto_type(cls, xml_node):
        try:
            network_object_type = xml_node.attrib[Attributes.XSI_NAMESPACE_TYPE]
        except KeyError:
            raise ValueError("XML Node is missing type attribute")
        try:
            return cls.registry[network_object_type].from_xml_node(xml_node)
        except KeyError:
            raise ValueError("Unknown network object type {}".format(network_object_type))

    @classmethod
    def from_st_network_object(cls, st_network_obj):
        raise NotImplementedError


class Service_Object(Base_Object, metaclass=SubclassWithIdentifierRegistry):
    def __init__(self, xml_tag, display_name, is_global, object_id, name, object_type, attr_type, application_id=None):
        super().__init__(xml_tag, display_name, is_global, object_id, name, object_type, application_id)
        self.set_attrib(Attributes.XSI_TYPE, attr_type)


    def as_service_type(self):
        raise NotImplementedError

    @staticmethod
    def from_xml_string_auto_type(xml_string):
        xml_node = ET.fromstring(xml_string)
        return Network_Object.from_xml_node_auto_type(xml_node)

    @classmethod
    def from_xml_node_auto_type(cls, xml_node):
        try:
            service_type = xml_node.attrib[Attributes.XSI_NAMESPACE_TYPE]
        except KeyError:
            raise ValueError("XML Node is missing type attribute")
        try:
            return cls.registry[service_type].from_xml_node(xml_node)
        except KeyError:
            raise ValueError("Unknown network object type {}".format(service_type))
